;CAML16Fd.S -- Forth Kernel
; CamelForth for the Zilog ZNEO (Z16F Series)
; portions (C) 2006 Douglas Beattie Jr.
;
; ..based on CamelForth for the Zilog Z80
; (C) 1994 Bradford J. Rodriguez
; Permission is granted to freely copy, modify,
; and distribute this program for personal or
; educational use.  Commercial inquiries should
; be directed to the author at 115 First St.,
; #105, Collingwood, Ontario L9Y 4W3, Canada
;
; CAML16Fd.S: CPU and Model Dependencies
;   Source code is for the ZDS-II ZNEO 4.10.2 macro
;   assembler (provided with Zilog Developer
;   Studio II).
;
;   Forth words are documented as follows:
;*   NAME     stack -- stack    description
;   Word names in upper case are from the ANS
;   Forth Core word set.  Names in lower case are
;   "internal" implementation words & extensions.
;
; Direct-Threaded Forth model for Zilog ZNEO
; 32 bit cell, 8 bit char, 8 bit (byte) address unit
; with typically 16-bit alignment.
; ===============================================

; ALIGNMENT AND PORTABILITY OPERATORS ===========
; Many of these are synonyms for other words,
; and so are defined as CODE words.

;C ALIGNED  addr -- a-addr       align given addr
    head ALIGNED,7,{"ALIGNED"},docode
        ;;ONEMINUS,LITERAL,3,AND,ONEPLUS
        DEC TOS
        OR TOS,#WORDSIZE-1
        INC TOS
        next

;C ALIGN    --                         align HERE
    head AALIGN,5,{"ALIGN"},docolon
;;noop:
    DL HERE,DUP,ALIGNED
    DL SWOP,MINUS,ALLOT,EXXIT

;Z CELL     -- n                 size of one cell
    head CELL,4,{"CELL"},docon
        DL CELLSIZE ;32-bit Cell = 4 byte

;C CELL+    a-addr1 -- a-addr2      add cell size
;   2 + ;
    head CELLPLUS,5,{"CELL+"},docode
        LEA TOS,CELLSIZE(TOS)
        next

;C CELLS    n1 -- n2            cells->adrs units
    head CELLS,5,{"CELLS"},docode
        LD WP,#4
        MUL TOS,WP
        next

;C CHAR+    c-addr1 -- c-addr2   add char size
    head CHARPLUS,5,{"CHAR+"},docolon
        DL ONEPLUS,EXXIT

;C CHARS    n1 -- n2            chars->adrs units
    head CHARS,5,{"CHARS"},docode
        next

;C >BODY    xt -- a-addr      adrs of param field
;   3 + ;                     Z80 (6 byte CALL)
    head TOBODY,5,{">BODY"},docolon
        DL LIT,6,PLUS,EXXIT  ;ZNEO (2-byte XT)


;X COMPILE,  xt --         append execution token
; I called this word ,XT before I discovered that
; it is defined in the ANSI standard as COMPILE,.
; On a DTC Forth this simply appends xt (like , )
; but on an STC Forth this must append 'CALL xt'.
    head COMMAXT,8,{"COMPILE,"},docode
        JP COMMA

;Z !CF    adrs cfa --   set code action of a word
;   0CD OVER C!         store 'CALL adrs' instr
;   1+ ! ;              Z80 VERSION
; Depending on the implementation this could
; append CALL adrs or JUMP adrs.
    head STORECF,3,{"!CF"},docolon
        ; ZNEO 'CALLA' opcode (0xF231) for DTC
        DL LIT,0F2h,OVER,CSTORE,ONEPLUS
        DL LIT,031h,OVER,CSTORE,ONEPLUS
        DL STORE,EXXIT

;Z ,CF    adrs --       append a code field
;   HERE !CF 3 ALLOT ;  Z80 VERSION (3 bytes)
    head COMMACF,3,{",CF"},docolon
        ;ZNEO DTC = 6 Bytes total (2-byte op, 4-byte addr)
        DL HERE,STORECF,LIT,6,ALLOT,EXXIT


;Z !COLON   --      change code field to docolon
;   -3 ALLOT docolon-adrs ,CF ;
; This should be used immediately after CREATE.
; This is made a distinct word, because on an STC
; Forth, colon definitions have no code field.
    head STORCOLON,6,{"!COLON"},docolon
        DL LIT,-6,ALLOT
        DL LIT,docolon,COMMACF,EXXIT



;Z ,EXXIT    --      append hi-level EXIT action
;   ['] EXIT ,XT ;
; This is made a distinct word, because on an STC
; Forth, it appends a RET instruction, not an xt.
    head CEXIT,5,{",EXIT"},docolon
        DL LIT,EXXIT,COMMAXT,EXXIT

; CONTROL STRUCTURES ============================
; These words allow Forth control structure words
; to be defined portably.

;Z ,BRANCH   xt --    append a branch instruction
; xt is the branch operator to use, e.g. qbranch
; or (loop).  It does NOT append the destination
; address.  On the Z80 this is equivalent to ,XT.
    head COMMABRANCH,7,{",BRANCH"},docode
        JP COMMA

;Z ,DEST   dest --        append a branch address
; This appends the given destination address to
; the branch instruction.  On the Z80 this is ','
; ...other CPUs may use relative addressing.
    head COMMADEST,5,{",DEST"},docode
        JP COMMA

;Z !DEST   dest adrs --    change a branch dest'n
; Changes the destination address found at 'adrs'
; to the given 'dest'.  On the Z80 this is '!'
; ...other CPUs may need relative addressing.
    head STOREDEST,5,{"!DEST"},docode
        JP STORE

; HEADER STRUCTURE ==============================
; The structure of the Forth dictionary headers
; (name, link, immediate flag, and "smudge" bit)
; does not necessarily differ across CPUs.  This
; structure is not easily factored into distinct
; "portable" words; instead, it is implicit in
; the definitions of FIND and CREATE, and also in
; NFA>LFA, NFA>CFA, IMMED?, IMMEDIATE, HIDE, and
; REVEAL.  These words must be (substantially)
; rewritten if either the header structure or its
; inherent assumptions are changed.


