;CAML16Fh.S -- Forth Kernel
; CamelForth for the Zilog ZNEO (Z16F Series)
; portions (C) 2006 Douglas Beattie Jr.
;
; ..based on CamelForth for the Zilog Z80
; (C) 1994 Bradford J. Rodriguez
; Permission is granted to freely copy, modify,
; and distribute this program for personal or
; educational use.  Commercial inquiries should
; be directed to the author at 115 First St.,
; #105, Collingwood, Ontario L9Y 4W3, Canada
;
;
; CAML16Fh.S: High Level Words
;   Source code is for the ZDS-II ZNEO 4.10.2 macro
;   assembler (provided with Zilog Developer
;   Studio II).
;
;   Forth words are documented as follows:
;*   NAME     stack -- stack    description
;   Word names in upper case are from the ANS
;   Forth Core word set.  Names in lower case are
;   "internal" implementation words & extensions.
; ===============================================

; SYSTEM VARIABLES & CONSTANTS ==================

;C BL      -- char            an ASCII space
    head BL,2,{"BL"},docon
        DL 20h

;Z tibsize  -- n         size of TIB
    head TIBSIZE,7,{"TIBSIZE"},docon
        DL SIZE_TIB-2  ; 2 chars safety zone

;X tib     -- a-addr     Terminal Input Buffer
;  HEX 82 CONSTANT TIB   CP/M systems: 126 bytes
;  HEX -80 USER TIB      others: below user area
    head TIB,3,{"TIB"},docon
TIB_pointer:
        DL InpBuffer

;Z u0      -- a-addr       current user area adrs
;  0 USER U0
    head U0,2,{"U0"},douser
        DL 0

;C >IN     -- a-addr        holds offset into TIB
;  2 USER >IN
    head TOIN,3,{">IN"},douser
        DL 1*CELLSIZE

;C BASE    -- a-addr       holds conversion radix
;  4 USER BASE
    head BASE,4,{"BASE"},douser
        DL 2*CELLSIZE

;C STATE   -- a-addr       holds compiler state
;  6 USER STATE
    head STATE,5,{"STATE"},douser
        DL 3*CELLSIZE

;Z dp      -- a-addr       holds dictionary ptr
;  8 USER DP
    head DP,2,{"DP"},douser
        DL 4*CELLSIZE

;Z 'source  -- a-addr      two cells: len, adrs
; 10 USER 'SOURCE
    head TICKSOURCE,7,{"\'SOURCE"},douser
        DL 5*CELLSIZE

;Z latest    -- a-addr     last word in dict.
;   14 USER LATEST
    head LATEST,6,{"LATEST"},douser
        DL 7*CELLSIZE

;Z hp       -- a-addr     HOLD pointer
;   16 USER HP
    head HP,2,{"HP"},douser
        DL 8*CELLSIZE

;Z LP       -- a-addr     Leave-stack pointer
;   18 USER LP
    head LP,2,{"LP"},douser
        DL 9*CELLSIZE

;Z l0       -- a-addr     bottom of Leave stack (grows upward)
;                         starts at the end of User Area
    head L0,2,{"L0"},douser
        DL SIZE_UA

;Z r0       -- a-addr     end of return stack (grows downward)
    head R_0,2,{"R0"},douser
        DL (SIZE_UA+SIZE_RSTK)

;X PAD       -- a-addr    user PAD buffer
;                         = end of hold area!
    head PAD,3,{"PAD"},douser
        DL (SIZE_UA+SIZE_RSTK+SIZE_HOLD)

;Z s0       -- a-addr     end of parameter stack
    head S0,2,{"S0"},douser
        DL (SIZE_UA+SIZE_RSTK+SIZE_HOLD+SIZE_PAD+SIZE_PSTK)-CELLSIZE
        ;minus CELLSIZE, since S! (SPSTORE) will pop a
		;non-existent TOS into register!

;Z uinit    -- addr  initial values for user area
    head UINIT,5,{"UINIT"},docreate
        DL 0,0,10,0     ; reserved,>IN,BASE,STATE
UDP     DL enddict      ; DP
        DL 0,0          ; SOURCE init'd elsewhere
ULATEST DL lastword     ; LATEST
        DL 0            ; HP init'd elsewhere
        DL 0            ; Leave Pointer init'd elsewhere

;Z #init    -- n    #bytes of user area init data
    head NINIT,5,{"#INIT"},docon
        ;;DL 18
        DL 10*CELLSIZE

; ARITHMETIC OPERATORS ==========================

;C S>D    n -- d          single -> double prec.
;   DUP 0< ;
    head STOD,3,{"S>D"},docolon
        DL DUP,ZEROLESS,EXXIT

;Z ?NEGATE  n1 n2 -- n3  negate n1 if n2 negative
;   0< IF NEGATE THEN ;        ...a common factor
    head QNEGATE,7,{"?NEGATE"},docolon
        DL ZEROLESS,QBRANCH,QNEG1,NEGATE
QNEG1:  DL EXXIT

;C ABS     n1 -- +n2     absolute value
;   DUP ?NEGATE ;
    head ABS,3,{"ABS"},docolon
        DL DUP,QNEGATE,EXXIT

;X DNEGATE   d1 -- d2     negate double precision
;   SWAP INVERT SWAP INVERT 1 M+ ;
    head DNEGATE,7,{"DNEGATE"},docolon
        DL SWOP,INVERT,SWOP,INVERT,LIT,1,MPLUS
        DL EXXIT

;Z ?DNEGATE  d1 n -- d2   negate d1 if n negative
;   0< IF DNEGATE THEN ;       ...a common factor
    head QDNEGATE,8,{"?DNEGATE"},docolon
        DL ZEROLESS,QBRANCH,DNEG1,DNEGATE
DNEG1:  DL EXXIT

;X DABS     d1 -- +d2    absolute value dbl.prec.
;   DUP ?DNEGATE ;
    head DABS,4,{"DABS"},docolon
        DL DUP,QDNEGATE,EXXIT

;C M*     n1 n2 -- d    signed 16*16->32 multiply
;   2DUP XOR >R        carries sign of the result
;   SWAP ABS SWAP ABS UM*
;   R> ?DNEGATE ;
    head MSTAR,2,{"M*"},docolon
        DL TWODUP,zXOR,TOR
        DL SWOP,ABS,SWOP,ABS,UMSTAR
        DL RFROM,QDNEGATE,EXXIT

;C SM/REM   d1 n1 -- n2 n3   symmetric signed div
;   2DUP XOR >R              sign of quotient
;   OVER >R                  sign of remainder
;   ABS >R DABS R> UM/MOD
;   SWAP R> ?NEGATE
;   SWAP R> ?NEGATE ;
; Ref. dpANS-6 section 3.2.2.1.
    head SMSLASHREM,6,{"SM/REM"},docolon
        DL TWODUP,zXOR,TOR,OVER,TOR
        DL ABS,TOR,DABS,RFROM,UMSLASHMOD
        DL SWOP,RFROM,QNEGATE,SWOP,RFROM,QNEGATE
        DL EXXIT

;C FM/MOD   d1 n1 -- n2 n3   floored signed div'n
;   DUP >R              save divisor
;   SM/REM
;   DUP 0< IF           if quotient negative,
;       SWAP R> +         add divisor to rem'dr
;       SWAP 1-           decrement quotient
;   ELSE R> DROP THEN ;
; Ref. dpANS-6 section 3.2.2.1.
    head FMSLASHMOD,6,{"FM/MOD"},docolon
        DL DUP,TOR,SMSLASHREM
        DL DUP,ZEROLESS,QBRANCH,FMMOD1
        DL SWOP,RFROM,PLUS,SWOP,ONEMINUS
        DL BRANCH,FMMOD2
FMMOD1: DL RFROM,DROP
FMMOD2: DL EXXIT

;C *      n1 n2 -- n3       signed multiply
;   M* DROP ;
    head STAR,1,{'*'},docolon
        DL MSTAR,DROP,EXXIT

;C /MOD   n1 n2 -- n3 n4    signed divide/rem'dr
;   >R S>D R> FM/MOD ;
    head SLASHMOD,4,{"/MOD"},docolon
        DL TOR,STOD,RFROM,FMSLASHMOD,EXXIT

;C /      n1 n2 -- n3       signed divide
;   /MOD nip ;
    head SLASH,1,{'/'},docolon
        DL SLASHMOD,NIP,EXXIT

;C MOD    n1 n2 -- n3       signed remainder
;   /MOD DROP ;
    head {MODD},3,{"MOD"},docolon
        DL SLASHMOD,DROP,EXXIT

;C */MOD  n1 n2 n3 -- n4 n5    n1*n2/n3, rem&quot
;   >R M* R> FM/MOD ;
    head SSMOD,5,{"*/MOD"},docolon
        DL TOR,MSTAR,RFROM,FMSLASHMOD,EXXIT

;C */     n1 n2 n3 -- n4        n1*n2/n3
;   */MOD nip ;
    head STARSLASH,2,{"*/"},docolon
        DL SSMOD,NIP,EXXIT

;C MAX    n1 n2 -- n3       signed maximum
;   2DUP < IF SWAP THEN DROP ;
    head MAX,3,{"MAX"},docolon
        DL TWODUP,LESS,QBRANCH,MAX1,SWOP
MAX1:   DL DROP,EXXIT

;C MIN    n1 n2 -- n3       signed minimum
;   2DUP > IF SWAP THEN DROP ;
    head MIN,3,{"MIN"},docolon
        DL TWODUP,GREATER,QBRANCH,MIN1,SWOP
MIN1:   DL DROP,EXXIT

; DOUBLE OPERATORS ==============================

;C 2@    a-addr -- x1 x2    fetch 2 cells
;   DUP CELL+ @ SWAP @ ;
;   the lower address will appear on top of stack
    head TWOFETCH,2,{"2@"},docolon
        DL DUP,CELLPLUS,FETCH,SWOP,FETCH,EXXIT

;C 2!    x1 x2 a-addr --    store 2 cells
;   SWAP OVER ! CELL+ ! ;
;   the top of stack is stored at the lower adrs
    head TWOSTORE,2,{"2!"},docolon
        DL SWOP,OVER,STORE,CELLPLUS,STORE,EXXIT

;C 2DROP  x1 x2 --          drop 2 cells
;   DROP DROP ;
    head TWODROP,5,{"2DROP"},docolon
        DL DROP,DROP,EXXIT

;C 2DUP   x1 x2 -- x1 x2 x1 x2   dup top 2 cells
;   OVER OVER ;
    head TWODUP,4,{"2DUP"},docolon
        DL OVER,OVER,EXXIT

;C 2SWAP  x1 x2 x3 x4 -- x3 x4 x1 x2  per diagram
;   ROT >R ROT R> ;
    head TWOSWAP,5,{"2SWAP"},docolon
        DL ROT,TOR,ROT,RFROM,EXXIT

;C 2OVER  x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2
;   >R >R 2DUP R> R> 2SWAP ;
    head TWOOVER,5,{"2OVER"},docolon
        DL TOR,TOR,TWODUP,RFROM,RFROM
        DL TWOSWAP,EXXIT

; INPUT/OUTPUT ==================================

;C PAGE      --   Clear display and home cursor (VT-100)
; { 1B 5B 32 4A } = ESC [2J
    head zPAGE,4,{"PAGE"},docolon
        DL XSQUOTE
        DB $F-$-1
        DB 27,"[H"  ;HOME CURSOR
        DB 27,"[2J" ;CLEAR DISPLAY
       .ALIGN WORDSIZE
$$:     DL TYPE,EXXIT

;C COUNT   c-addr1 -- c-addr2 u  counted->adr/len
;   DUP CHAR+ SWAP C@ ;
    head COUNT,5,{"COUNT"},docolon
        DL DUP,CHARPLUS,SWOP,CFETCH,EXXIT

;C CR      --               output newline
;   0D EMIT 0A EMIT ;
    head CR,2,{"CR"},docolon
        DL LIT,0Dh,EMIT,LIT,0Ah,EMIT,EXXIT

;C SPACE   --               output a space
;   BL EMIT ;
    head zSPACE,5,{"SPACE"},docolon
        DL BL,EMIT,EXXIT

;C SPACES   n --            output n spaces
;   BEGIN DUP WHILE SPACE 1- REPEAT DROP ;
    head SPACES,6,{"SPACES"},docolon
SPCS1:  DL DUP,QBRANCH,SPCS2
        DL zSPACE,ONEMINUS,BRANCH,SPCS1
SPCS2:  DL DROP,EXXIT

;Z umin     u1 u2 -- u      unsigned minimum
;   2DUP U> IF SWAP THEN DROP ;
    head UMIN,4,{"UMIN"},docolon
        DL TWODUP,UGREATER,QBRANCH,UMIN1,SWOP
UMIN1:  DL DROP,EXXIT

;Z umax    u1 u2 -- u       unsigned maximum
;   2DUP U< IF SWAP THEN DROP ;
    head UMAX,4,{"UMAX"},docolon
        DL TWODUP,ULESS,QBRANCH,UMAX1,SWOP
UMAX1:  DL DROP,EXXIT

;C ACCEPT  c-addr +n -- +n'  get line from term'l
;   OVER + 1- OVER      -- sa ea a
;   BEGIN KEY           -- sa ea a c
;   DUP 0D <> WHILE
;       DUP EMIT        -- sa ea a c
;       DUP 8 = IF  DROP 1-    >R OVER R> UMAX
;             ELSE  OVER C! 1+ OVER UMIN
;       THEN            -- sa ea a
;   REPEAT              -- sa ea a c
;   DROP NIP SWAP - ;
    head ACCEPT,6,{"ACCEPT"},docolon
        DL OVER,PLUS,ONEMINUS,OVER
ACC1:   DL KEY,DUP,LIT,0Dh,NOTEQUAL,QBRANCH,ACC5
        DL DUP,EMIT,DUP,LIT,8,zEQUAL,QBRANCH,ACC3
        DL DROP,ONEMINUS,TOR,OVER,RFROM,UMAX
        DL BRANCH,ACC4
ACC3:   DL OVER,CSTORE,ONEPLUS,OVER,UMIN
ACC4:   DL BRANCH,ACC1
ACC5:   DL DROP,NIP,SWOP,MINUS,EXXIT

;C TYPE    c-addr +n --     type line to term'l
;   ?DUP IF
;     OVER + SWAP DO I C@ EMIT LOOP
;   ELSE DROP THEN ;
    head TYPE,4,{"TYPE"},docolon
        DL QDUP,QBRANCH,TYP4
        DL OVER,PLUS,SWOP,XDO
TYP3:   DL II,CFETCH,EMIT,XLOOP,TYP3
        DL BRANCH,TYP5
TYP4:   DL DROP
TYP5:   DL EXXIT

;Z (S")     -- c-addr u   run-time code for S"
;   R> COUNT 2DUP + ALIGNED >R  ;
    head XSQUOTE,4,{"(S\")"},docolon
        DL RFROM,COUNT,TWODUP,PLUS,ALIGNED,TOR
        DL EXXIT

;C S"       --         compile in-line string
;   COMPILE (S")  [ HEX ]
;   22 WORD C@ 1+ ALIGNED ALLOT ; IMMEDIATE
    immed SQUOTE,2,{"S\""},docolon
        DL LIT,XSQUOTE,COMMAXT
        DL LIT,22H,zWORD,CFETCH,ONEPLUS
        DL ALIGNED,ALLOT,EXXIT

;C ."       --         compile string to print
;   POSTPONE S"  POSTPONE TYPE ; IMMEDIATE
    immed DOTQUOTE,2,{".\""},docolon
        DL SQUOTE
        DL LIT,TYPE,COMMAXT
        DL EXXIT

; NUMERIC OUTPUT ================================
; Numeric conversion is done l.s.digit first, so
; the output buffer is built backwards in memory.

; Some double-precision arithmetic operators are
; needed to implement ANSI numeric conversion.

;Z UD/MOD   ud1 u2 -- u3 ud4   32/16->32 divide
;   >R 0 R@ UM/MOD  ROT ROT R> UM/MOD ROT ;
    head UDSLASHMOD,6,{"UD/MOD"},docolon
        DL TOR,LIT,0,RFETCH,UMSLASHMOD,ROT,ROT
        DL RFROM,UMSLASHMOD,ROT,EXXIT

;Z UD*      ud1 d2 -- ud3      32*16->32 multiply
;   DUP >R UM* DROP  SWAP R> UM* ROT + ;
    head UDSTAR,3,{"UD*"},docolon
        DL DUP,TOR,UMSTAR,DROP
        DL SWOP,RFROM,UMSTAR,ROT,PLUS,EXXIT

;C HOLD  char --        add char to output string
;   -1 HP +!  HP @ C! ;
    head HOLD,4,{"HOLD"},docolon
        DL LIT,-1,HP,PLUSSTORE
        DL HP,FETCH,CSTORE,EXXIT

;C <#    --             begin numeric conversion
;   PAD HP ! ;          (initialize Hold Pointer)
    head LESSNUM,2,{"<#"},docolon
        DL PAD,HP,STORE,EXXIT

;Z >digit   n -- c      convert to 0..9A..Z
;   [ HEX ] DUP 9 > 7 AND + 30 + ;
    head TODIGIT,6,{">DIGIT"},docolon
        DL DUP,LIT,9,GREATER,LIT,7,zAND,PLUS
        DL LIT,30H,PLUS,EXXIT

;C #     ud1 -- ud2     convert 1 digit of output
;   BASE @ UD/MOD ROT >digit HOLD ;
    head NUM,1,{'#'},docolon
        DL BASE,FETCH,UDSLASHMOD,ROT,TODIGIT
        DL HOLD,EXXIT

;C #S    ud1 -- ud2     convert remaining digits
;   BEGIN # 2DUP OR 0= UNTIL ;
    head NUMS,2,{"#S"},docolon
NUMS1:  DL NUM,TWODUP,zOR,ZEROEQUAL,QBRANCH,NUMS1
        DL EXXIT

;C #>    ud1 -- c-addr u    end conv., get string
;   2DROP HP @ PAD OVER - ;
    head NUMGREATER,2,{"#>"},docolon
        DL TWODROP,HP,FETCH,PAD,OVER,MINUS,EXXIT

;C SIGN  n --           add minus sign if n<0
;   0< IF 2D HOLD THEN ;
    head SIGN,4,{"SIGN"},docolon
        DL ZEROLESS,QBRANCH,SIGN1,LIT,2DH,HOLD
SIGN1:  DL EXXIT

;C U.    u --           display u unsigned
;   <# 0 #S #> TYPE SPACE ;
    head UDOT,2,{"U."},docolon
        DL LESSNUM,LIT,0,NUMS,NUMGREATER,TYPE
        DL zSPACE,EXXIT

;C .     n --           display n signed
;   <# DUP ABS 0 #S ROT SIGN #> TYPE SPACE ;
    head DOT,1,{'.'},docolon
        DL LESSNUM,DUP,ABS,LIT,0,NUMS
        DL ROT,SIGN,NUMGREATER,TYPE,zSPACE,EXXIT

;Z Z.    u --           display u unsigned with no trailing space
;   <# 0 #S #> TYPE ;
    head ZDOT,2,"Z.",docolon
        DL LESSNUM,LIT,0,NUMS,NUMGREATER,TYPE
        DL EXXIT


;C DECIMAL  --      set number base to decimal
;   10 BASE ! ;
    head DECIMAL,7,{"DECIMAL"},docolon
        DL LIT,10,BASE,STORE,EXXIT

;X HEX     --       set number base to hex
;   16 BASE ! ;
    head HEX,3,{"HEX"},docolon
        DL LIT,16,BASE,STORE,EXXIT

; DICTIONARY MANAGEMENT =========================

;C HERE    -- addr      returns dictionary ptr
;   DP @ ;
    head HERE,4,{"HERE"},docolon
        DL DP,FETCH,EXXIT

;C ALLOT   n --         allocate n bytes in dict
;   DP +! ;
    head ALLOT,5,{"ALLOT"},docolon
        DL DP,PLUSSTORE,EXXIT

; Note: , and C, are only valid for combined
; Code and Data spaces.

;C ,    x --           append cell to dict
;   HERE ! 1 CELLS ALLOT ;
    head COMMA,1,2Ch,docolon
        DL HERE,STORE,LIT,1,CELLS,ALLOT,EXXIT

;C C,   char --        append char to dict
;   HERE C! 1 CHARS ALLOT ;
    head CCOMMA,2,{"C,"},docolon
        DL HERE,CSTORE,LIT,1,CHARS,ALLOT,EXXIT

; INTERPRETER ===================================
; Note that NFA>LFA, NFA>CFA, IMMED?, and FIND
; are dependent on the structure of the Forth
; header.  This may be common across many CPUs,
; or it may be different.

;C SOURCE   -- adr n    current input buffer
;   'SOURCE 2@ ;        length is at lower adrs
    head SOURCE,6,{"SOURCE"},docolon
        DL TICKSOURCE,TWOFETCH,EXXIT

;X /STRING  a u n -- a+n u-n   trim string
;   ROT OVER + ROT ROT - ;
    head SLASHSTRING,7,{"/STRING"},docolon
        DL ROT,OVER,PLUS,ROT,ROT,MINUS,EXXIT

;Z >counted  src n dst --     copy to counted str
;   2DUP C! CHAR+ SWAP CMOVE ;
    head TOCOUNTED,8,{">COUNTED"},docolon
        DL TWODUP,CSTORE,CHARPLUS,SWOP,CMOVE,EXXIT

;C WORD   char -- c-addr n   word delim'd by char
;   DUP  SOURCE >IN @ /STRING   -- c c adr n
;   DUP >R   ROT SKIP           -- c adr' n'  ...... '>R PUTS WHAT!?
;   OVER >R  ROT SCAN           -- adr" n"    ...... '>R PUTS WHAT!?
;   DUP IF CHAR- THEN        skip trailing delim.
;   R> R> ROT -   >IN +!        update >IN offset
;   TUCK -                      -- adr' N
;   HERE >counted               --
;   HERE                        -- a
;   BL OVER COUNT + C! ;    append trailing blank
    head zWORD,4,{"WORD"},docolon
        DL DUP,SOURCE,TOIN,FETCH,SLASHSTRING
        DL DUP,TOR,ROT,SKIP
        DL OVER,TOR,ROT,SCAN
        DL DUP,QBRANCH,WORD1,ONEMINUS  ; char-
WORD1:  DL RFROM,RFROM,ROT,MINUS,TOIN,PLUSSTORE
        DL TUCK,MINUS
        DL HERE,TOCOUNTED,HERE
        DL BL,OVER,COUNT,PLUS,CSTORE,EXXIT

;Z NFA>LFA   nfa -- lfa    name adr -> link field
;   3 - ;
    head NFATOLFA,7,{"NFA>LFA"},docolon
        DL LIT,2,CELLS,MINUS,EXXIT

;Z NFA>CFA   nfa -- cfa    name adr -> code field
;   COUNT 7F AND + ;       mask off 'smudge' bit
    head NFATOCFA,7,{"NFA>CFA"},docolon
        ; start of string, 7-bit count
        DL COUNT,LIT,07Fh,zAND,PLUS
        DL ALIGNED,EXXIT


;Z IMMED?    nfa -- f      fetch immediate flag
;   1- C@ ;                     nonzero if immed
    head IMMEDQ,6,{"IMMED?"},docolon
        DL ONEMINUS,CFETCH,EXXIT

;C FIND   c-addr -- c-addr 0   if not found
;C                  xt  1      if immediate
;C                  xt -1      if "normal"
;   LATEST @ BEGIN             -- a nfa
;       2DUP OVER C@ CHAR+     -- a nfa a nfa n+1
;       S=                     -- a nfa f
;       DUP IF
;           DROP
;           NFA>LFA @ DUP      -- a link link
;       THEN
;   0= UNTIL                   -- a nfa  OR  a 0
;   DUP IF
;       NIP DUP NFA>CFA        -- nfa xt
;       SWAP IMMED?            -- xt iflag
;       0= 1 OR                -- xt 1/-1
;   THEN ;
    head FIND,4,{"FIND"},docolon
        DL BREAK
        DL LATEST,FETCH
FIND1:  DL TWODUP,OVER,CFETCH,CHARPLUS
        DL SEQUAL,DUP,QBRANCH,FIND2
        DL DROP,NFATOLFA,FETCH,DUP
FIND2:  DL ZEROEQUAL,QBRANCH,FIND1
        DL DUP,QBRANCH,FIND3
        DL NIP,DUP,NFATOCFA
        DL SWOP,IMMEDQ,ZEROEQUAL,LIT,1,zOR
FIND3:  DL EXXIT

;C LITERAL  x --        append numeric literal
;   STATE @ IF ['] LIT ,XT , THEN ; IMMEDIATE
; This tests STATE so that it can also be used
; interpretively.  (ANSI doesn't require this.)
    immed LITERAL,7,{"LITERAL"},docolon
        DL STATE,FETCH,QBRANCH,LITER1
        DL LIT,LIT,COMMAXT,COMMA
LITER1: DL EXXIT

;Z DIGIT?   c -- n -1   if c is a valid digit
;Z            -- x  0   otherwise
;   [ HEX ] DUP 39 > 100 AND +     silly looking
;   DUP 140 > 107 AND -   30 -     but it works!
;   DUP BASE @ U< ;
    head DIGITQ,6,{"DIGIT?"},docolon
        DL DUP,LIT,39H,GREATER,LIT,100H,zAND,PLUS
        DL DUP,LIT,140H,GREATER,LIT,107H,zAND
        DL MINUS,LIT,30H,MINUS
        DL DUP,BASE,FETCH,ULESS,EXXIT

;Z ?SIGN   adr n -- adr' n' f  get optional sign
;Z  advance adr/n if sign; return NZ if negative
;   OVER C@                 -- adr n c
;   2C - DUP ABS 1 = AND    -- +=-1, -=+1, else 0
;   DUP IF 1+               -- +=0, -=+2
;       >R 1 /STRING R>     -- adr' n' f
;   THEN ;
    head QSIGN,5,{"?SIGN"},docolon
        DL OVER,CFETCH,LIT,2CH,MINUS,DUP,ABS
        DL LIT,1,zEQUAL,zAND,DUP,QBRANCH,QSIGN1
        DL ONEPLUS,TOR,LIT,1,SLASHSTRING,RFROM
QSIGN1: DL EXXIT

;C >NUMBER  ud adr u -- ud' adr' u'
;C                      convert string to number
;   BEGIN
;   DUP WHILE
;       OVER C@ DIGIT?
;       0= IF DROP EXIT THEN
;       >R 2SWAP BASE @ UD*
;       R> M+ 2SWAP
;       1 /STRING
;   REPEAT ;
    head TONUMBER,7,{">NUMBER"},docolon
TONUM1: DL DUP,QBRANCH,TONUM3
        DL OVER,CFETCH,DIGITQ
        DL ZEROEQUAL,QBRANCH,TONUM2,DROP,EXXIT
TONUM2: DL TOR,TWOSWAP,BASE,FETCH,UDSTAR
        DL RFROM,MPLUS,TWOSWAP
        DL LIT,1,SLASHSTRING,BRANCH,TONUM1
TONUM3: DL EXXIT

;Z ?NUMBER  c-addr -- n -1      string->number
;Z                 -- c-addr 0  if convert error
;   DUP  0 0 ROT COUNT      -- ca ud adr n
;   ?SIGN >R  >NUMBER       -- ca ud adr' n'
;   IF   R> 2DROP 2DROP 0   -- ca 0   (error)
;   ELSE 2DROP NIP R>
;       IF NEGATE THEN  -1  -- n -1   (ok)
;   THEN ;
    head QNUMBER,7,{"?NUMBER"},docolon
        DL DUP,LIT,0,DUP,ROT,COUNT
        DL QSIGN,TOR,TONUMBER,QBRANCH,QNUM1
        DL RFROM,TWODROP,TWODROP,LIT,0
        DL BRANCH,QNUM3
QNUM1:  DL TWODROP,NIP,RFROM,QBRANCH,QNUM2,NEGATE
QNUM2:  DL LIT,-1
QNUM3:  DL EXXIT

;Z INTERPRET    i*x c-addr u -- j*x
;Z                      interpret given buffer
; This is a common factor of EVALUATE and QUIT.
; ref. dpANS-6, 3.4 The Forth Text Interpreter
;   'SOURCE 2!  0 >IN !
;   BEGIN
;   BL WORD DUP C@ WHILE        -- textadr
;       FIND                    -- a 0/1/-1
;       ?DUP IF                 -- xt 1/-1
;           1+ STATE @ 0= OR    immed or interp?
;           IF EXECUTE ELSE ,XT THEN
;       ELSE                    -- textadr
;           ?NUMBER
;           IF POSTPONE LITERAL     converted ok
;           ELSE COUNT TYPE 3F EMIT CR ABORT  err
;           THEN
;       THEN
;   REPEAT DROP ;
    head INTERPRET,9,{"INTERPRET"},docolon
        DL TICKSOURCE,TWOSTORE,LIT,0,TOIN,STORE
INTER1: DL BL,zWORD,DUP,CFETCH,QBRANCH,INTER9
        DL FIND,QDUP,QBRANCH,INTER4
        DL ONEPLUS,STATE,FETCH,ZEROEQUAL,zOR
        DL QBRANCH,INTER2
        DL EXECUTE,BRANCH,INTER3
INTER2: DL COMMAXT
INTER3: DL BRANCH,INTER8
INTER4: DL QNUMBER,QBRANCH,INTER5
        DL LITERAL,BRANCH,INTER6
INTER5: DL COUNT,TYPE,LIT,3FH,EMIT,CR,ABORT
INTER6:
INTER8: DL BRANCH,INTER1
INTER9: DL DROP,EXXIT

;C EVALUATE  i*x c-addr u -- j*x  interprt string
;   'SOURCE 2@ >R >R  >IN @ >R
;   INTERPRET
;   R> >IN !  R> R> 'SOURCE 2! ;
    head EVALUATE,8,{"EVALUATE"},docolon
        DL TICKSOURCE,TWOFETCH,TOR,TOR
        DL TOIN,FETCH,TOR,INTERPRET
        DL RFROM,TOIN,STORE,RFROM,RFROM
        DL TICKSOURCE,TWOSTORE,EXXIT

;C QUIT     --    R: i*x --    interpret from kbd
;   L0 LP !  R0 RP!   0 STATE !
;   BEGIN
;       TIB DUP TIBSIZE ACCEPT  SPACE
;       INTERPRET
;       STATE @ 0= IF CR ." OK" THEN
;   AGAIN ;
    head QUIT,4,{"QUIT"},docolon
        ;;;DL INIT_UART
   ;;DL BREAK ;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        DL L0,LP,STORE
        DL R_0,RPSTORE,LIT,0,STATE,STORE
QUIT1:  DL TIB,DUP,TIBSIZE,ACCEPT,zSPACE
        DL INTERPRET
        DL STATE,FETCH,ZEROEQUAL,QBRANCH,QUIT2
        DL CR,XSQUOTE
        DB 3,"OK "
       .ALIGN WORDSIZE
        DL TYPE
QUIT2:  DL BRANCH,QUIT1

;C ABORT    i*x --   R: j*x --   clear stk & QUIT
;   S0 SP!  QUIT ;
    head ABORT,5,{"ABORT"},docolon
  ;;DL BREAK
        DL S0,SPSTORE,QUIT   ; QUIT never returns

;Z ?ABORT   f c-addr u --      abort & print msg
;   ROT IF TYPE ABORT THEN 2DROP ;
    head QABORT,6,{"?ABORT"},docolon
        DL ROT,QBRANCH,QABO1,TYPE,ABORT
QABO1:  DL TWODROP,EXXIT

;C ABORT"  i*x 0  -- i*x   R: j*x -- j*x  x1=0
;C         i*x x1 --       R: j*x --      x1<>0
;   POSTPONE S" POSTPONE ?ABORT ; IMMEDIATE
    immed ABORTQUOTE,6,{"ABORT\""},docolon
        DL SQUOTE
        DL LIT,QABORT,COMMAXT
        DL EXXIT

;C '    -- xt           find word in dictionary
;   BL WORD FIND
;   0= ABORT" ?" ;
    head TICK,1,{"\'"},docolon
        DL BL,zWORD,FIND,ZEROEQUAL,XSQUOTE
        DB 1,'?'
       .ALIGN WORDSIZE
        DL QABORT,EXXIT

;C CHAR   -- char           parse ASCII character
;   BL WORD 1+ C@ ;
    head CHAR,4,{"CHAR"},docolon
        DL BL,zWORD,ONEPLUS,CFETCH,EXXIT

;C [CHAR]   --          compile character literal
;   CHAR  ['] LIT ,XT  , ; IMMEDIATE
    immed BRACCHAR,6,{"[CHAR]"},docolon
        DL CHAR
        DL LIT,LIT,COMMAXT
        DL COMMA,EXXIT

;C (    --                     skip input until )
;   [ HEX ] 29 WORD DROP ; IMMEDIATE
    immed PAREN,1,{'('},docolon
        DL LIT,29H,zWORD,DROP,EXXIT

;Z \    --       skip remainder of line
;  1 WORD DROP ; IMMEDIATE
    immed BACKSLASH,1,{'\\'},docolon
        DL LIT,1,zWORD,DROP,EXXIT


; COMPILER ======================================

;C CREATE   --      create an empty definition
;   LATEST @ , 0 C,         link & immed field
;   HERE LATEST !           new "latest" link
;   BL WORD C@ 1+ ALLOT         name field
;   docreate ,CF                code field
    head CREATE,6,{"CREATE"},docolon
;;        DL LATEST,FETCH,COMMA,LIT,0,CCOMMA
        DL LATEST,FETCH,COMMA,LIT,0,COMMA
        DL HERE,LATEST,STORE
        DL BL,zWORD,CFETCH,ONEPLUS,ALLOT
 DL AALIGN
        DL LIT,docreate,COMMACF,EXXIT

;Z (DOES>)  --      run-time action of DOES>
;   R>              adrs of headless DOES> def'n
;   LATEST @ NFA>CFA    code field to fix up
;   !CF ;
    head XDOES,7,{"(DOES>)"},docolon
        DL RFROM,LATEST,FETCH,NFATOCFA,STORECF
        DL EXXIT

;C DOES>    --      change action of latest def'n
;   COMPILE (DOES>)
;   dodoes ,CF ; IMMEDIATE
    immed DOES,5,{"DOES>"},docolon
        DL LIT,XDOES,COMMAXT
        DL LIT,dodoes,COMMACF,EXXIT

;C RECURSE  --      recurse current definition
;   LATEST @ NFA>CFA ,XT ; IMMEDIATE
    immed RECURSE,7,{"RECURSE"},docolon
        DL LATEST,FETCH,NFATOCFA,COMMAXT,EXXIT

;C [        --      enter interpretive state
;   0 STATE ! ; IMMEDIATE
    immed LEFTBRACKET,1,{'['},docolon
        DL LIT,0,STATE,STORE,EXXIT

;C ]        --      enter compiling state
;   -1 STATE ! ;
    head RIGHTBRACKET,1,{']'},docolon
        DL LIT,-1,STATE,STORE,EXXIT

;Z HIDE     --      "hide" latest definition
;   LATEST @ DUP C@ 80 OR SWAP C! ;
    head HIDE,4,{"HIDE"},docolon
        DL LATEST,FETCH,DUP,CFETCH,LIT,80H,zOR
        DL SWOP,CSTORE,EXXIT

;Z REVEAL   --      "reveal" latest definition
;   LATEST @ DUP C@ 7F AND SWAP C! ;
    head REVEAL,6,{"REVEAL"},docolon
        DL LATEST,FETCH,DUP,CFETCH,LIT,7FH,zAND
        DL SWOP,CSTORE,EXXIT

;C IMMEDIATE   --   make last def'n immediate
;   1 LATEST @ 1- C! ;   set immediate flag
    head IMMEDIATE,9,{"IMMEDIATE"},docolon
        DL LIT,1,LATEST,FETCH,ONEMINUS,CSTORE
        DL EXXIT

;C :        --      begin a colon definition
;   CREATE HIDE ] !COLON ;
    head COLON,1,{':'},docode
        CALL docolon    ; code fwd ref explicitly
        DL CREATE,HIDE,RIGHTBRACKET,STORCOLON
        DL EXXIT

;C ;
;   REVEAL  ,EXXIT
;   POSTPONE [  ; IMMEDIATE
    immed SEMICOLON,1,{';'},docolon
        DL REVEAL,CEXIT
        DL LEFTBRACKET,EXXIT

;C [']  --         find word & compile as literal
;   '  ['] LIT ,XT  , ; IMMEDIATE
; When encountered in a colon definition, the
; phrase  ['] xxx  will cause   LIT,xxt  to be
; compiled into the colon definition (where
; (where xxt is the execution token of word xxx).
; When the colon definition executes, xxt will
; be put on the stack.  (All xt's are one cell.)
    immed BRACTICK,3,{"[\']"},docolon
        DL TICK               ; get xt of 'xxx'
        DL LIT,LIT,COMMAXT    ; append LIT action
        DL COMMA,EXXIT         ; append xt literal

;C POSTPONE  --   postpone compile action of word
;   BL WORD FIND
;   DUP 0= ABORT" ?"
;   0< IF   -- xt  non immed: add code to current
;                  def'n to compile xt later.
;       ['] LIT ,XT  ,      add "LIT,xt,COMMAXT"
;       ['] ,XT ,XT         to current definition
;   ELSE  ,XT      immed: compile into cur. def'n
;   THEN ; IMMEDIATE
    immed POSTPONE,8,{"POSTPONE"},docolon
        DL BL,zWORD,FIND,DUP,ZEROEQUAL,XSQUOTE
        DB 1,'?'
       .ALIGN WORDSIZE
        DL QABORT,ZEROLESS,QBRANCH,POST1
        DL LIT,LIT,COMMAXT,COMMA
        DL LIT,COMMAXT,COMMAXT,BRANCH,POST2
POST1:  DL COMMAXT
POST2:  DL EXXIT

;Z COMPILE   --   append inline execution token
;   R> DUP CELL+ >R @ ,XT ;
; The phrase ['] xxx ,XT appears so often that
; this word was created to combine the actions
; of LIT and ,XT.  It takes an inline literal
; execution token and appends it to the dict.
;    head COMPILE,{"COMPILE"},docolon
;        DW RFROM,DUP,CELLPLUS,TOR
;        DW FETCH,COMMAXT,EXXIT
; N.B.: not used in the current implementation

; CONTROL STRUCTURES ============================

;C IF       -- adrs    conditional forward branch
;   ['] qbranch ,BRANCH  HERE DUP ,DEST ;
;   IMMEDIATE
    immed IFF,2,{"IF"},docolon
        DL LIT,QBRANCH,COMMABRANCH
        DL HERE,DUP,COMMADEST,EXXIT

;C THEN     adrs --        resolve forward branch
;   HERE SWAP !DEST ; IMMEDIATE
    immed THEN,4,{"THEN"},docolon
        DL HERE,SWOP,STOREDEST,EXXIT

;C ELSE     adrs1 -- adrs2    branch for IF..ELSE
;   ['] branch ,BRANCH  HERE DUP ,DEST
;   SWAP  POSTPONE THEN ; IMMEDIATE
    immed ELLSE,4,{"ELSE"},docolon
        DL LIT,BRANCH,COMMABRANCH
        DL HERE,DUP,COMMADEST
        DL SWOP,THEN,EXXIT

;C BEGIN    -- adrs        target for bwd. branch
;   HERE ; IMMEDIATE
    immed BEGIN,5,{"BEGIN"},docode
        JP HERE

;C UNTIL    adrs --   conditional backward branch
;   ['] qbranch ,BRANCH  ,DEST ; IMMEDIATE
;   conditional backward branch
    immed UNTIL,5,{"UNTIL"},docolon
        DL LIT,QBRANCH,COMMABRANCH
        DL COMMADEST,EXXIT

;X AGAIN    adrs --      uncond'l backward branch
;   ['] branch ,BRANCH  ,DEST ; IMMEDIATE
;   unconditional backward branch
    immed AGAIN,5,{"AGAIN"},docolon
        DL LIT,BRANCH,COMMABRANCH
        DL COMMADEST,EXXIT

;C WHILE    -- adrs         branch for WHILE loop
;   POSTPONE IF ; IMMEDIATE
    immed WHILE,5,{"WHILE"},docode
        JP IFF

;C REPEAT   adrs1 adrs2 --     resolve WHILE loop
;   SWAP POSTPONE AGAIN POSTPONE THEN ; IMMEDIATE
    immed REPEAT,6,{"REPEAT"},docolon
        DL SWOP,AGAIN,THEN,EXXIT

;Z >L   x --   L: -- x        move to leave stack
;   CELL LP +!  LP @ ! ;      (L stack grows up)
    head TOL,2,{">L"},docolon
        DL CELL,LP,PLUSSTORE,LP,FETCH,STORE,EXXIT

;Z L>   -- x   L: x --      move from leave stack
;   LP @ @  CELL NEGATE LP +! ;
    head LFROM,2,{"L>"},docolon
        DL LP,FETCH,FETCH
        DL CELL,NEGATE,LP,PLUSSTORE,EXXIT

;C DO       -- adrs   L: -- 0
;   ['] xdo ,XT   HERE     target for bwd branch
;   0 >L ; IMMEDIATE           marker for LEAVEs
    immed DO,2,{"DO"},docolon
        DL LIT,XDO,COMMAXT,HERE
        DL LIT,0,TOL,EXXIT

;Z ENDLOOP   adrs xt --   L: 0 a1 a2 .. aN --
;   ,BRANCH  ,DEST                backward loop
;   BEGIN L> ?DUP WHILE POSTPONE THEN REPEAT ;
;                                 resolve LEAVEs
; This is a common factor of LOOP and +LOOP.
    head ENDLOOP,7,{"ENDLOOP"},docolon
        DL COMMABRANCH,COMMADEST
LOOP1:  DL LFROM,QDUP,QBRANCH,LOOP2
        DL THEN,BRANCH,LOOP1
LOOP2:  DL EXXIT

;C LOOP    adrs --   L: 0 a1 a2 .. aN --
;   ['] xloop ENDLOOP ;  IMMEDIATE
    immed LOOP,4,{"LOOP"},docolon
        DL LIT,XLOOP,ENDLOOP,EXXIT

;C +LOOP   adrs --   L: 0 a1 a2 .. aN --
;   ['] xplusloop ENDLOOP ;  IMMEDIATE
    immed PLUSLOOP,5,{"+LOOP"},docolon
        DL LIT,XPLUSLOOP,ENDLOOP,EXXIT

;C LEAVE    --    L: -- adrs
;   ['] UNLOOP ,XT
;   ['] branch ,BRANCH   HERE DUP ,DEST  >L
;   ; IMMEDIATE      unconditional forward branch
    immed LEAVE,5,{"LEAVE"},docolon
        DL LIT,UNLOOP,COMMAXT
        DL LIT,BRANCH,COMMABRANCH
        DL HERE,DUP,COMMADEST,TOL,EXXIT

; OTHER OPERATIONS ==============================

;X WITHIN   n1|u1 n2|u2 n3|u3 -- f   n2<=n1<n3?
;  OVER - >R - R> U< ;          per ANS document
    head WITHIN,6,{"WITHIN"},docolon
        DL OVER,MINUS,TOR,MINUS,RFROM,ULESS,EXXIT

;C MOVE    addr1 addr2 u --     smart move
;             VERSION FOR 1 ADDRESS UNIT = 1 CHAR
;  >R 2DUP SWAP DUP R@ +     -- ... dst src src+n
;  WITHIN IF  R> CMOVE>        src <= dst < src+n
;       ELSE  R> CMOVE  THEN ;          otherwise
    head MOVE,4,{"MOVE"},docolon
        DL TOR,TWODUP,SWOP,DUP,RFETCH,PLUS
        DL WITHIN,QBRANCH,MOVE1
        DL RFROM,CMOVEUP,BRANCH,MOVE2
MOVE1:  DL RFROM,CMOVE
MOVE2:  DL EXXIT

;C DEPTH    -- +n        number of items on stack
;   SP@ S0 SWAP - 2/ ;   32-BIT VERSION, BASED ON 'CELLSIZE'
    head DEPTH,5,{"DEPTH"},docolon
        DL SPFETCH,S0,SWOP,MINUS
        DL LIT,CELLSIZE,SLASH,EXXIT

;C ENVIRONMENT?  c-addr u -- false   system query
;                         -- i*x true
;   2DROP 0 ;       the minimal definition!
    head ENVIRONMENTQ,12,{"ENVIRONMENT?"},docolon
        DL TWODROP,LIT,0,EXXIT

; UTILITY WORDS AND STARTUP =====================

;X WORDS    --          list all words in dict.
;   LATEST @ BEGIN
;       DUP COUNT TYPE SPACE
;       NFA>LFA @
;   DUP 0= UNTIL
;   DROP ;
    head WORDS,5,{"WORDS"},docolon
        DL LATEST,FETCH
WDS1:   DL DUP,COUNT,TYPE,zSPACE,NFATOLFA,FETCH
        DL DUP,ZEROEQUAL,QBRANCH,WDS1
        DL DROP,EXXIT

;X .S      --           print stack contents
;   SP@ S0 - IF
;       SP@ S0 2 - DO I @ U. -2 +LOOP
;   THEN ;
    head DOTS,2,{".S"},docolon
        DL SPFETCH,S0,MINUS,QBRANCH,DOTS2
        DL SPFETCH,S0,LIT,CELLSIZE,MINUS,XDO
DOTS1:  DL II,FETCH,UDOT,LIT,-CELLSIZE,XPLUSLOOP,DOTS1
DOTS2:  DL EXXIT

;Z COLD     --      cold start Forth system
;   UINIT U0 #INIT CMOVE      init user area
;   80 COUNT INTERPRET       interpret CP/M cmd
;   ." Z80 CamelForth etc."
;   ABORT ;
    head COLD,4,{"COLD"},docolon
        ;;;;DL INIT_UART            ;initialize UART for terminal I/O
        DL UINIT,U0,NINIT,CMOVE     ;initialize User area
;;      DW LIT,InpBuffer,COUNT,INTERPRET
;;      DW LIT,0398h,TMR1_INIT      ;turn on 1-msec timer
        DL XSQUOTE      ; sign-on banner to console
        DB $F-$+1,0Dh,0Ah,0Dh,0Ah
        DB "Camel FORTH for ZiLOG ZNEO v1.0, 21 Oktober 2006"
$$:     DB 0Dh,0Ah
       .ALIGN WORDSIZE
       DL TYPE,ABORT       ; ABORT never returns


